%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Function: hcc_extra
%
% Authors: Ansel Y. Rodrguez Gonzlez 1,2 (Team Leader) 
%          Samantha Barajas            1,3
%          Ramn Aranda                1,2
%          Yoan Martnez Lpez         4
%          Julio Madera                4
%
% 1 Unidad de Transferencia Tecnolgica Tepic del Centro de Investigacin 
%   Cientfica y de Educacin Superior de Ensenada, Mxico
%
% 2 Consejo Nacional de Ciencia y Tecnologa, Mxico 
%
% 3 Universidad Autnoma de Nayarit, Mxico
%
% 4 Universidad de Camagey, Cuba
%
% Modified by AYRG \CICESE 13/06/2021

function [Fit_and_p,FVr_bestmemit, fitMaxVector] = ...
    hcc_extra(hc2RCEDUMDAparameters,caseStudyData,otherParameters,low_habitat_limit,up_habitat_limit, ~, FVr_bestmemit, fitMaxVector, gen, evalCount, I_repeatEvals, FullStoreSol, FullStoreVal, FullStoreCount)

    maxEvals      = hc2RCEDUMDAparameters.maxEvals;
    popSize       = 1;
    ratioFactor   = hc2RCEDUMDAparameters.I_finalHC_ratioFactor;
    maxStep       = (1 / (hc2RCEDUMDAparameters.I_CodeSize-1)) * ratioFactor; 
        
     

    bestSol = FVr_bestmemit;
    bestVal = fitMaxVector(1,gen);
    
    previuosSol = bestSol;
    previuosVal = bestVal;
    
       
    while evalCount < maxEvals - popSize
        
        Pop = generateNewPop(bestSol,bestVal,previuosSol,previuosVal,popSize,maxStep,low_habitat_limit, up_habitat_limit);        
        
        [newVal, evalCount] = my_feval(otherParameters.fnc,Pop,caseStudyData, otherParameters, evalCount,I_repeatEvals);
        
        [newBestVal,newBestIndex] = min(newVal);  
        
        fprintf('Fitness value: %f %f\n',bestVal, newBestVal);
        fprintf('Generation: %d\n',gen);
        
        gen = gen + 1;  
        
        if (newBestVal<bestVal)
            previuosSol         = bestSol;
            previuosVal         = bestVal;
            
            bestSol             = Pop(newBestIndex,:);
            bestVal             = newBestVal;            
        end
        
        fitMaxVector(1,gen) = bestVal;        
        
    end
   
   FVr_bestmemit = bestSol;
   Fit_and_p=[fitMaxVector(1,gen) 0];
   
function Pop = generateNewPop(bestSol,bestVal,previousSol,previuosVal,popSize,maxStep,low_habitat_limit, up_habitat_limit)
    [~,cws] = size(bestSol);    
    Pop   = rand(popSize,cws);
    
    for j=1:cws
        
       if (previuosVal == bestVal)
          Pop(1,j) =  bestSol(j) + ( (0.5 - Pop(1,j))  * maxStep * (up_habitat_limit(j) - low_habitat_limit(j))); 
       else
         if (bestSol(j) == previousSol(j))
            Pop(1,j) =  bestSol(j) + ( (0.5 - Pop(1,j))  * maxStep * (up_habitat_limit(j) - low_habitat_limit(j))); 
         end
         if (bestSol(j) > previousSol(j))
            Pop(1,j) =  bestSol(j) + ( (0.5 - Pop(1,j))  * maxStep * (up_habitat_limit(j) - low_habitat_limit(j))); 
         end
         if (bestSol(j) < previousSol(j))
            Pop(1,j) =  bestSol(j) + ( (0.5 - Pop(1,j))  * maxStep * (up_habitat_limit(j) - low_habitat_limit(j))); 
         end         
       end
                     
       if (Pop(1,j)>up_habitat_limit(j)) 
           Pop(1,j) = up_habitat_limit(j);
       end 
       if (Pop(1,j)<low_habitat_limit(j)) 
           Pop(1,j) = low_habitat_limit(j);
       end
    end
end
    
function [Val, evalCount] = my_feval(fnc,deCodedPop,caseStudyData, otherParameters, evalCount, I_repeatEvals)    
    switch I_repeatEvals
        case 0
            %%%% Removed method. It is not relevant for HC2RCEDUMDA
           % [Val, evalCount] = my_feval_withStore(fnc,deCodedPop,caseStudyData, otherParameters, evalCount);
        case 2
           [Val, evalCount] = my_feval_withStoreForElitism(fnc,deCodedPop,caseStudyData, otherParameters, evalCount); 
        otherwise    
            [s,~] = size(deCodedPop);    
            evalCount = evalCount + s;
            [Val, ~] = feval(fnc,deCodedPop,caseStudyData, otherParameters);
            fprintf('evalCount: %d\n',evalCount);       
    end
end  

function [Val, evalCount] = my_feval_withStoreForElitism(fnc,deCodedPop,caseStudyData, otherParameters, evalCount)    
     [cs,ws] = size(deCodedPop);
     Val    = nan(1,cs);
     
     for iii=1:cs
        index = searchSolution(deCodedPop(iii,:),ws);
        if index > 0
            Val(iii) = FullStoreVal(index);           
        else
           auxPop = deCodedPop(iii,:);
           [auxVal, ~] = feval(fnc,auxPop,caseStudyData, otherParameters);
           Val(iii) = auxVal;
           evalCount = evalCount + 1;          
        end
     end    

     fprintf('evalCount: %d\n',evalCount);
end     
      
function index = searchSolution(sol, ws)
     index = 0; 
    
     for iii=1:FullStoreCount                 
         flag = 1;
         for jjj=1:ws
            if FullStoreSol(iii,jjj)~=sol(jjj)
              flag = 0;
              break;
            end
         end
         
         if flag == 1
            index = iii;
            fprintf('*');
            break; 
         end         
     end
end
     
     
end   
